/*
	MSCDEX Class interface declaration
	Copyright (c) 1994~1995 Marco Savegnago
*/

#ifndef	__CDROM_H
#define	__CDROM_H

#include "tbobject.h"
#include "trigitem.h"
#include "taction.h"

#include <mem.h>

	/* Addressing modes */
#define	ADDR_HSG	0
#define	ADDR_RED	1

	/* Device driver commands */
#define	DEVRDIOCTL	 3	/* IOCTL read			*/
#define	DEVWRIOCTL	 12	/* IOCTL write			*/

#define	DEVPLAY		132	/* Device Play			*/
#define	DEVSTOP		133	/* Stop device play		*/

	/* CDROM Device IOCTL commands */
#define	IOI_ret_addr		0
#define	IOI_loc_head		1
#define	IOI_io_query		2
#define	IOI_err_stats		3
#define	IOI_audio_info		4
#define	IOI_rd_drv_bytes	5
#define	IOI_dev_status		6
#define	IOI_ret_sectsize	7
#define	IOI_ret_volsize		8
#define	IOI_media_changed	9
#define	IOI_audio_diskinfo	10
#define	IOI_audio_trackinfo	11
#define	IOI_audio_qchaninfo	12
#define	IOI_audio_subinfo	13
#define	IOI_upc_code		14
#define	IOI_cmd_max		14

#define	IOO_eject_disc		0
#define	IOO_lock_door		1
#define	IOO_reset_drv		2
#define	IOO_set_audio_param	3
#define	IOO_wr_drv_bytes	4
#define	IOO_cmd_max			4
#define IOO_close_tray      5

#define	error_drive_not_ready	21

/*
Status

The status word also has the same format as described in the MS-PRM. It is 0 on entry and is set by the device driver.

 	Bit 15		- Error bit
 	Bit 14-10	- Reserved
	Bit  9		- Busy
 	Bit  8		- Done
	Bit  7-0	- Error code (bit 15 on)

Bit 15, the error bit, is set by the device driver if an error is detected or if an invalid request is made to the driver. The low 8 bits indicate the error code.

Bit 9, the busy bit, should be set by the device driver when the drive is in audio play mode. Device drivers should fail all requests to the physical device that require head movement when the device is playing and return the request with this bit and the error bit set and an error code. Requests that would not interrupt audio play may return without error but will also have this bit set when the drive is in audio play mode. Play mode can be terminated prematurely with a reset or STOP AUDIO request and a ne

w request can be made at that point. Monitoring this bit in each successive request, an Audio Q-Channel Info IOCTL for example, will tell when play mode is complete.

Bit 8, the done bit, is set by the device driver when the operation is finished.

Error codes are the following:

	0	Write-protect violation
	1	Unknown unit
	2	Drive not ready
	3	Unknown command
	4	CRC error
	5	Bad drive request structure length
	6	Seek error
	7	Unknown media
	8	Sector not found
	9	Printer out of paper
	A	Write fault
	B	Read fault
	C	General failure
	D	Reserved
	E	Reserved
    F   Invalid disk change
*/


/* Device driver request header */
struct RequestHdr
{
	byte		rqhLen;
	byte		rqhUnit;
	byte		rqhCmd;

    union
    {
        struct
        {
            byte    errorCode   : 8;
            byte    done        : 1;
            byte    busy        : 1;
			byte    reserved    : 5;
			byte    error       : 1;
        }   rqhStatusDetails;

        word        rqhStatus;
    };

    byte        rqhRsvd[8];
};

	/* Request header for INIT function */
struct InitHdr
{
	RequestHdr	initRqh;
	byte		initUnits;
	byte	far	*initEndaddr;
	byte	far	*initBpbarr;
	byte		initDevno;
};

/* Request header for IOCTL command */
struct IOCTLHeader
{
	RequestHdr	ioctlRqh;
	byte		ioctlMedia;
	byte far*	ioctlXfer;
	word		ioctlNbytes;
	word		ioctlSector;
	byte far*	ioctlVolId;
};

/* Record for audio_diskinfo IOCTL call */
struct DiskInfoRec
{
	byte	cmdCode;
	byte	lowTNO;
	byte	highTNO;
	dword	leadOut;
	DiskInfoRec();
};

inline 	DiskInfoRec::DiskInfoRec()
{
	cmdCode = IOI_audio_diskinfo;
	lowTNO	= highTNO = 0;
	leadOut	= 0L;
}

/*
- Device Status
________________

DevStat		DB	6		; Control block code
		DD	?		; Device parameters
________________

The device driver will return a 32-bit value. Bit 0 is the least significant bit. The bits are interpreted as follows:

Bit 0		0	Door closed
		1	Door open
Bit 1		0	Door locked
		1	Door unlocked
Bit 2		0	Supports only cooked reading
		1	Supports cooked and raw reading
Bit 3		0	Read only
		1	Read/write
Bit 4		0	Data read only
		1	Data read and plays audio/video tracks
Bit 5		0	No interleaving
		1	Supports ISO-9660 interleaving using interleave size 
			and skip factor
Bit 6		0	Reserved
Bit 7		0	No prefetching
		1	Supports prefetching requests
Bit 8		0	No audio channel manipulation
		1	Supports audio channel manipulation
Bit 9		0	Supports HSG addressing mode
		1	Supports HSG and Red Book addressing modes
Bit 10		0	Reserved
Bit 11		0	Disc is present in drive
		1	No disc is present in drive
Bit 12		0	Doesn't support R-W sub-channels
		1	Supports R-W sub-channels
Bit 13-31   0   Reserved (all 0)
*/


/* Record for device status IOCTL call */
struct DeviceStatusRec
{
	byte        cmdCode;
	Boolean     doorOpen        : 1;
	Boolean     doorUnlocked    : 1;
	Boolean     cookedAndRaw    : 1;
	Boolean     readWrite       : 1;
	Boolean     dataAndAudio    : 1;
	Boolean     interleaving    : 1;
	Boolean     reserved1       : 1;
	Boolean     prefetching     : 1;
	Boolean     audioManip      : 1;
	Boolean     RedBooksupport  : 1;
	Boolean     reserved2       : 1;
	Boolean     noDiskPresent   : 1;
	Boolean     supportRWSubCh  : 1;
	byte        reserved3       : 3;
	byte        reserved4[2];

	DeviceStatusRec();
};

inline DeviceStatusRec::DeviceStatusRec()
{
	memset(this, 0, sizeof(*this));
	cmdCode	= IOI_dev_status;
}

/* 	Struttura che descrive il buffer per la richiesta di eject del disco	*/

struct  EjectDiskRec
{
	byte	cmdCode;
	EjectDiskRec();
};

inline EjectDiskRec::EjectDiskRec()
{
	cmdCode =	IOO_eject_disc;
}

/* 	Struttura che descrive il buffer per la richiesta di chiusura del cassetto	*/

struct  CloseTrayRec
{
	byte	cmdCode;
	CloseTrayRec();
};

inline CloseTrayRec::CloseTrayRec()
{
	cmdCode =	IOO_close_tray;
}

/* 	Struttura che descrive il buffer per la richiesta di reset del drive	*/

struct  ResetDiskRec
{
	byte	cmdCode;
	ResetDiskRec();
};

inline ResetDiskRec::ResetDiskRec()
{
	cmdCode =	IOO_reset_drv;
}

/* 	Struttura che descrive il buffer per la richiesta di lock/unlock del cassetto	*/

struct  LockUnlockDoorRec
{
	byte	cmdCode;
	byte	action;
	LockUnlockDoorRec();
};

inline LockUnlockDoorRec::LockUnlockDoorRec()
{
	cmdCode	=	IOO_lock_door;
	action  = 	0;
}

/* Format for CDROM device driver header at CS:0 of device driver
** Slightly extended from standard MSDOS character device driver
*/

struct DeviceHeader
{
	struct DeviceHeader far* sdevnext;	/* Pointer to next dev header */
	word	  sdevatt;					/* Attributes of the device   */
	void	  (near *sdevstrat)();		/* Strategy entry point       */
	void	  (near *sdevint)();		/* Interrupt entry point      */
	byte	  sdevname[8];				/* Name of device (only first byte used for block) */
	word	  sdevrsvd;					/* Reserved word	      	  */
	byte	  sdevlet;					/* Drive letter of first unit */
	byte	  sdevunits;				/* Number of units handled    */
};

/* Record format for information returned with Get CDROM drive
** letter device list function request
*/
struct DeviceList
{
	byte		subUnit;
	DeviceHeader far*	devAddr;
};

/* Record for audio_trackinfo IOCTL call */

struct TNOInfoRec
{
	byte	cmdCode;
	byte	tno;
	dword	startAddr;
	byte	ctrl;

	TNOInfoRec();
};

inline 	TNOInfoRec::TNOInfoRec()
{
	cmdCode 	= IOI_audio_trackinfo;
	tno			= 0;
	startAddr   = 0L;
	ctrl		= 0;
}

	/* Request header for Read/Write command */
struct ReadWriteLHdr
{
	RequestHdr	rwlRqh;
	byte		rwlAddrmd;
	byte far*	rwlXfer;
	word		rwlNunSects;
	dword		rwlSectNo;
	byte		rwlMode;
	byte		rwlILSize;
	byte		rwlILSkip;
	word		rwlReqno;
};

	/* Request header for Play Audio command */
struct PlayReqHdr
{
	RequestHdr	plRqh;
	byte		plAddrmd;
	dword		plStart;
	dword		plNum;
};

	/* Record for audio_qchaninfo IOCTL call */
struct QChanInfoRec
{
	byte	cmdCode;
	byte	ctrl;
	byte	tno;
	byte	x;
	byte	min;
	byte	sec;
	byte	frame;
	byte	zero;
	byte	pmin;
	byte	psec;
	byte	pframe;

	QChanInfoRec();
};

inline 	QChanInfoRec::QChanInfoRec()
{
	memset(this, 0, sizeof(*this));
	cmdCode = IOI_audio_qchaninfo;
}

/*
- Audio Channel Info
________________

AudInfo	DB	4		; Control block code
		DB	?		; Input channel (0, 1, 2, or 3) for output channel 0
		DB	?		; Volume control (0 - 0xff) for output channel 0
		DB	?		; Input channel (0, 1, 2, or 3) for output channel 1
		DB	?		; Volume control (0 - 0xff) for output channel 1
		DB	?		; Input channel (0, 1, 2, or 3) for output channel 2
		DB	?		; Volume control (0 - 0xff) for output channel 2
		DB	?		; Input channel (0, 1, 2, or 3) for output channel 3
		DB	?		; Volume control (0 - 0xff) for output channel 3
________________
*/

struct	InputAndVolumeOfAutoChannelInfo
{
	byte	input;
	byte	volume;
};

struct	AudioChannelInfoRec
{
	byte							cmdCode;
	InputAndVolumeOfAutoChannelInfo channel[4];

	AudioChannelInfoRec();
};

inline	AudioChannelInfoRec::AudioChannelInfoRec()
{
	memset(this, 0, sizeof(*this));
}

class	CDROM
:
	public TTriggerItem
{
protected:
	Boolean	isInPlay;	//	Flag vero se sono in play
	word	id;			//	ID del disco

	DeviceHeader 		far*	devAddr;			// Puntatore al device driver
	DiskInfoRec	 		far*	diskInfo;			// Buffer che contiene le diskinfo
	TNOInfoRec 	 		far* 	tnoInfo;			// Buffer temporaneo usato per leggere le track info
	QChanInfoRec 		far*	qChanInfoRec;		// Buffer per richiedere le info sullo stato del cd
	DeviceStatusRec 	far*	deviceStatus;		// Buffer per richiedere il device status
	EjectDiskRec		far*	ejectDiskRec;		// Buffer per richiedere l'eject del disco
	CloseTrayRec		far*	closeTrayRec;		// Buffer per richiedere la chiusura del cassetto
	ResetDiskRec		far*	resetDiskRec;		// Buffer per richiedere il reset del drive
	LockUnlockDoorRec   far*	lockUnlockDoorRec;  // Buffer per richiedere il blocco e lo sblocco del cassetto
	AudioChannelInfoRec far*	audioChannelInfoRec;// Buffer per richiedere le info sulle traccie audio;

	static		IOCTLHeader ioctl;
	static		PlayReqHdr	playReq;
public:

	CDROM(word id_, DeviceHeader far*	devAddr_);	// Costruttore
	virtual	~CDROM();								// Distruttore

	TBNSCollection	*tracks;				// Puntatore alla lista di tracce

	word 	lowTrack(); 					// Ritorna il numero della traccia inferiore
	word	highTrack();					// Ritorna il numero della traccia superiore
	dword	leadOut();						// Ritorna la fine assolura del disco

	dword	getBeginOfDisk();				// Ritorna l'inizio assoluto del disco
	dword	getLenOfDisk();					// Ritorna la lunghezza assoluta del disco

	void 	printAddress(char* str, dword	addr);

	dword	getTrackStartREDAddress(word track);
	dword	getTrackLenRED(word track);
	dword 	red2hsg(dword l);

	/*	Membri per ridefinire le funzionalita' di TTriggerItem	*/

	virtual	Boolean Is() ;		//	Ritorna vero se e' in play
	virtual	Boolean IsNot() ;	//	Ritorna vero se non e' in play
	virtual	void	Reset();	//	Resetta la condizione

	Boolean	isPlay();			//	Ritorna vero se il drive e' in play
	Boolean	isDoorOpen();		//	Ritorna vero se il cassetto e' aperto
	Boolean	isDoorLocked();		//	Ritorna vero se il cassetto e' bloccato
	Boolean isDiskInDrive();	//	Ritorna vero se il c'e' un disco

	void	eject();					//	Espelle il disco
	void	closeTray();				//  Chiude il cassetto
	void	resetDrive();				//  Resetta il drive
	void	lockDoor();					//	Lockka il cassetto
	void	unlockDoor();				//  Sblokka il cassetto
	word	getTrackCount();			//  Ritorna il numero di traccie

	Boolean	readTOC();					//	Legge il contenuto del disco
	Boolean stop();						//  Ferma la riproduzione del disco
	Boolean play();						//  Avvia la riproduzione dell'intero disco
	Boolean play(word track);			//	Avvia la riproduzione del singolo brano del disco
	void	checkStatus();				//  Verifica lo stato del device


	Boolean	setInput(byte channel, byte input); 	// 	Imposta l'input source per il canale
	Boolean	setVolume(byte channel, byte volume);	//  Imposta il volume per il canale

	byte	getInput(byte channel);		//	Ritorna l'input associato al canale
	byte	getVolume(byte channel);	//	Ritorna il volume associato al canale

protected:
	Boolean play(dword start, dword len);

	Boolean IOCTLInputCmdAudioDiskInfo();	//	IOCTL Input cmds
	Boolean IOCTLInputCmdDeviceStatus();
	Boolean IOCTLInputCmdAudioTrackInfo();
	Boolean IOCTLInputCmdAudioChannelInfo();

	Boolean IOCTLOutputCmdEjectDisk();		// 	IOCTL	Output  cmds
	Boolean IOCTLOutputCmdCloseTray();
	Boolean IOCTLOutputCmdReset();
	Boolean IOCTLOutputLockDoor(Boolean lock);
	Boolean IOCTLOutputCmdAudioChannelInfo();

	void	sendReq(void far * buffer);		//	Invia la richiesta IOCTL al driver

protected:
};

inline	Boolean	CDROM::isDoorOpen()
{
	return	deviceStatus->doorOpen;
}

inline 	Boolean	CDROM::isDoorLocked()
{
	return 	Boolean(deviceStatus->doorUnlocked == False);
}

inline 	Boolean CDROM::isDiskInDrive()
{
	return 	Boolean(deviceStatus->noDiskPresent == False);
}

inline 	word	CDROM::lowTrack()
{
	return diskInfo->lowTNO;
}

inline	word	CDROM::highTrack()
{
	return diskInfo->highTNO;
}

inline dword	CDROM::leadOut()
{
	return diskInfo->leadOut;
}

inline	Boolean	CDROM::isPlay()
{
	return 	isInPlay;
}

inline Boolean CDROM::Is()
{
	return Boolean(isInPlay == True);
}

inline Boolean CDROM::IsNot()
{
	return Boolean(isInPlay == False);
}

inline void	CDROM::Reset()
{
}

inline word	CDROM::getTrackCount()
{
	return tracks->getCount();
}

inline	void	CDROM::eject()
{
	IOCTLOutputCmdEjectDisk();
}

inline	void	CDROM::closeTray()
{
	IOCTLOutputCmdCloseTray();
}

inline	void	CDROM::resetDrive()
{
	IOCTLOutputCmdReset();
}

inline	void	CDROM::lockDoor()
{
	IOCTLOutputLockDoor(True);
}

inline	void	CDROM::unlockDoor()
{
	IOCTLOutputLockDoor(False);
}

inline	void	CDROM::checkStatus()
{
	IOCTLInputCmdDeviceStatus();
}

// 	Imposta l'input source per il canale

inline	Boolean	CDROM::setInput(byte channel, byte input)
{
	IOCTLInputCmdAudioChannelInfo();
	audioChannelInfoRec->channel[channel].input = input;
	return	IOCTLOutputCmdAudioChannelInfo();
}

//  Imposta il volume per il canale

inline	Boolean	CDROM::setVolume(byte channel, byte volume)
{
	IOCTLInputCmdAudioChannelInfo();
	audioChannelInfoRec->channel[channel].volume = volume;
	return	IOCTLOutputCmdAudioChannelInfo();
}
//  Ritorna l'input del canale

inline	byte	CDROM::getInput(byte channel)
{
	IOCTLInputCmdAudioChannelInfo();
	return audioChannelInfoRec->channel[channel].input;
}
//  Ritorna il valore del volume del canale

inline byte	CDROM::getVolume(byte channel)
{
	IOCTLInputCmdAudioChannelInfo();
	return audioChannelInfoRec->channel[channel].volume;
}

#define	MAXCDROMDEVICE	8

/*
	La classe CDROMManager serve per gestire un insieme di CDROM Drives
	raggruppandoli all'interno di una collezione e fornendo le funzionalita'
	per accedere a ogni singolo CD o tutti insieme

	Se questo non fosse necessario, per accedere alla singola unita' e'
	sufficente richiederne il puntatore al CDROMManager e poi gestirla
	direttamente con le funzioni di CDROM
*/


class	CDROMManager
:
	public TAction
{
protected:

	byte		MSCDEXMinorVersion;		// Versione minore di MSCDEX rilevato
	byte		MSCDEXMajorVersion;		// Versione maggiore di MSCDEX rilevato

	static		DeviceList 	DeviceTable[MAXCDROMDEVICE];

	word		NumOfDrives;			// Numero di drives rilevati
	char		FirstDriveLetter;		// Prima lettera del primo CDROM rilevato

	TBNSCollection	*cdroms;			// Collezione di CDROM

public:

	CDROMManager();						// Ctor
	virtual	~CDROMManager();			// Dtor

	Boolean		FindCDRomDrives();		// Cerca i CDROM drives
	Boolean 	ReadCDRomsTOC();		// Legge le TOC di tutti i CDROMS

	void		ResetDrives();			// Resetta tutti i drives
	void		ResetCD(word cd);		// Resetta tutti i drives

	Boolean 	ReadCDROMS();			// Legge il contenuto (brani) del CDROM
	void        freeCDROMS();			// Cancella la collezione di CDROM

	void		StopAll();						// Ferma la riproduzione di tutti i CD
	void		PlayAll();						// Avvia la riproduzione di tutti i CD

	Boolean 	StopCD(word cd);				// Ferma la riproduzione del CD

	Boolean 	PlayCD(word cd);				// Avvia la riproduzione completa del singolo CD
	Boolean 	PlayCD(word cd, word track);	// Avvia la riproduzione del singolo CD alla singola traccia


	void		EjectAll();				// Apre il cassetto a tutti i CDROM
	void		EjectCD(word cd);		// Apre il cassetto del singolo CDROM

	void		UnlockAll();
	void		LockAll();

	void		CloseTrayAll();			// Chiude tutti i cassetti dei drives
	void		CloseTrayCD(word cd);	// Chiude il cassetto del drive specificato

	CDROM*		getCDROM(word id);	//  Ritorna il puntatore al CDROM

	word		getNumOfDrives();	//	Ritorna il numero di drives rilevati

	void		checkStatus();

	byte		getVolume(word cd, byte channel);	//	Ritorna il volume del CANALE del CD
	byte		getInput(word cd, byte channel);	//	Ritorna l'input del CANALE del CD

	void		setAllVolume(byte volume);			//  Imposta tutti i volumi

	Boolean		setVolume(word cd, byte channel, byte volume);	// 	Imposta il volume del canale del CD
	void		setVolume(word cd, byte volume);

	Boolean		getInput(word cd, word channel, byte input);	//  Imposta l'input del canale del CD

	void		getCDStart(char* str, word cd);					//	Ritorna la stringa formattata con il tempo di start del CD
	void		getTrackStart(char* str, word cd, word track);	//	Ritorna la stringa formattata con il tempo di start del brano del CD

	void		getCDLen(char* str, word cd);					//	Ritorna la stringa formattata con la lunghezza formattata del brano del CD
	void		getTrackLen(char* str, word cd, word track);	//	Ritorna la stringa formattata con la lunghezza del brano del CD

	//	TAction member

	// Boolean	action(word pos, Boolean set, void* value ...);
	virtual	Boolean	action(word pos, Boolean set, int	value ...);
};

/*
	Questa funzione e' da preferire all'accesso diretto all'array
	per consentire di gestire l'eccezione nel caso in cui il CDROM
	richiesto non sia presente nella lista
*/

inline	CDROM* CDROMManager::getCDROM(word id)
{
	if(id < cdroms->getCount())
		return	(CDROM*) cdroms->at(id);

    return NULL;
}


inline	word	CDROMManager::getNumOfDrives()
{
	return NumOfDrives;
}

#endif
