/*
	Base Object implementation
	(c) Copyright 1992/1994 Marco Savegnago IW3FQG
*/

// local include

#include "tbobject.h"

//	CRTL include

#include <mem.h>
#include <assert.h>

TBNSCollection::TBNSCollection( ccIndex aLimit, ccIndex aDelta ) :
	count( 0 ),
    items( 0 ),
    limit( 0 ),
    delta( aDelta ),
    shouldDelete( True )
{
	setLimit( aLimit );
}

TBNSCollection::TBNSCollection() :
    count( 0 ),
    items( 0 ),
    limit( 0 ),
    delta( 0 ),
    shouldDelete( True )
{
    items = 0;
}

TBNSCollection::~TBNSCollection()
{
    delete [] items;
}

void TBNSCollection::shutDown()
{
    if( shouldDelete )
        freeAll();
    setLimit(0);
	TBobject::shutDown();
}

void *TBNSCollection::at( ccIndex index )
{
	if( index < 0 || index >= count )
		error(1,index);
    return items[index];
}

void TBNSCollection::atRemove( ccIndex index )
{
	if( index < 0 || index >= count )
		error(1,index);

    count--;
    memmove( &items[index], &items[index+1], (count-index)*sizeof(void *) );
}

void TBNSCollection::atFree( ccIndex index )
{
    void *item = at( index );
    atRemove( index );
    freeItem( item );
}

void TBNSCollection::atInsert(ccIndex index, void *item)
{
    if( index < 0 )
		error(1,index);
    if( count == limit )
        setLimit(count + delta);

    memmove( &items[index+1], &items[index], (count-index)*sizeof(void *) );
	count++;

    items[index] = item;
}

void TBNSCollection::atPut( ccIndex index, void *item )
{
    if( index >= count )
		error(1,index);

    items[index] = item;
}

void TBNSCollection::remove( void *item )
{
    atRemove( indexOf(item) );
}

void* TBNSCollection::remove(word itemNo)
{
	assert( count );

    void* first=items[itemNo];
	atRemove( itemNo );

    return first;
}

void TBNSCollection::removeAll()
{
    count = 0;
}

#pragma argsused
void TBNSCollection::error( ccIndex code, ccIndex index)
{
	assert(0);
}

void *TBNSCollection::firstThat( ccTestFunc Test, void *arg )
{
	for( ccIndex i = 0; i < count; i++ )
	{
		if( Test( items[i], arg ) == True )
			return items[i];
	}

	return 0;
}

void *TBNSCollection::lastThat( ccTestFunc Test, void *arg )
{
	for( ccIndex i = count; i > 0; i-- )
	{
		if( Test( items[i-1], arg ) == True )
			return items[i-1];
	}
	return 0;
}

void TBNSCollection::forEach( ccAppFunc action, void *arg )
{
	for( ccIndex i = 0; i < count; i++ )
		action( items[i], arg );
}

void TBNSCollection::free( void *item )
{
	remove( item );
    freeItem( item );
}

void TBNSCollection::freeAll()
{
	for( ccIndex i =  0; i < count; i++ )
        freeItem( at(i) );
    count = 0;
}

void TBNSCollection::freeItem( void *item )
{
    delete item;
}

#pragma warn -rvl
ccIndex TBNSCollection::indexOf(void *item)
{
	for( ccIndex i = 0; i < count; i++ )
		if( item == items[i] )
			return i;

	error(1,0);
}
#pragma warn +rvl

ccIndex TBNSCollection::insert( void *item )
{
	ccIndex loc = count;
    atInsert( count, item );
    return loc;
}

void TBNSCollection::pack()
{
    void **curDst = items;
	void **curSrc = items;
	void **last = items + count;
	while( curSrc < last )
	{
		if( *curSrc != 0 )
			*curDst++ = *curSrc;
		*curSrc++;
	}
}

void TBNSCollection::setLimit(ccIndex aLimit)
{
	if( aLimit < count )
		aLimit =  count;
	if( aLimit > maxCollectionSize)
		aLimit = maxCollectionSize;

	if( aLimit != limit )
	{
		void **aItems;
		if (aLimit == 0 )
			aItems = 0;
		else
		{
			aItems = new void *[aLimit];
			if( count !=  0 )
				memcpy( aItems, items, count*sizeof(void *) );
		}
		delete [] items;
		items =  aItems;
		limit =  aLimit;
	}
}
